/*
 BiPhase_rcv.v data receiver - for use with BiPhase_xmit.v

 Kirk Weedman

 May 16, 2012
 
 Concept:
   Receive data across N wires between FPGAs using a BiPhase Mark data encoding similar to the AES/SPDIF format

 BiPhase Mark Encoding:

   Data:        1   0   0   1   0   1   0
	           +-+ +---+   +-+ +---+ +-+   +-
	BPM:      -+ +-+   +---+ +-+   +-+ +---+
    
                1 1 1 0 0 0 1 0
              +-----+     +-+ +
	Preamble: -+     +-----+ +-+
    Preamble uses 8 time slots and each data bit 2 time slots

 Method:
   OSL (Output Stream Length): Number of Data bits that will be transmitted in the data stream after the Preamble
   BIT_TIME: Amount of time it takes to transmit a Data "logic 0" or logic 1" (2 time slots)
	DEAD_TIME: time between transmissions.

    After reset, the DIN signal is held low for DLY_TIME by the transmitter.  The transmitter then sends the
    following sequence:  Preamble, Data stream, DEAD_TIME ...

                 +---------------+ 
     data bit ---+               +-  "logic 0" if signal was low at end of last BIT_TIME
                 |<--BIT_TIME--->|
 
              ---+               +-
     data bit    +---------------+   "logic 0" if signal was high at end of last BIT_TIME
                 |<--BIT_TIME--->|

                 +-------+       +-  
     data bit ---+       +-------+   "logic 1" if signal was low at end of last BIT_TIME
                 |<--BIT_TIME--->|
 
              ---+       +-------+
     data bit    +-------+       +-  "logic 1" if signal was high at end of last BIT_TIME
                 |<--BIT_TIME--->|


                           +-----------------------+                +---
  DIN  --------------------+-----------------------+----------------+---
       |<---- DLY_TIME---->|<--- OSL*BIT_TIME ---->|<-- DEAD_TIME ->|
                           |<----------- 1/SEND_FREQ -------------->|

   BIT_TIME must be determined by this receiver by sampling the Preamble information.  If P1 is equal to
   the time of the logic 1's, then BIT_TIME = P1*2/3.  If P2 is equal to the time of the logic 0's, then
   P1 and P2 should match within 12.5%.  If the final 1 0 slot sequence of the preamble is not within 12.5%
   of the BIT_TIME, then an error is assumed and the logic will continue to search for the preamble sequence.
   12.5% is chosen as it's 1/8th which is an easy binary calculation by using a right shift of 3 bits.
*/

`timescale 1ns/100ps

module BiPhase_rcv (irst, iclk, xrst, xclk, xrcv_data, xrcv_rdy, xrcv_ack, din);

parameter OSL          = 32;  // Output Stream Length
parameter OSW          = 1;   // Output Stream Width
parameter SLOWEST_FREQ = 10000;
parameter ICLK_FREQ    = 144000000;
parameter XCLK_FREQ    = 48000000;
parameter TPD          = 1;   // only for simulation purposes

localparam DATA_BITS   = OSL * OSW;       // Total number of data bits (logic 0 or 1) to receive: must be > 5?
localparam DC          = clogb2(OSL-1);
localparam DLY_TIME    = ICLK_FREQ/SLOWEST_FREQ;
localparam DS          = clogb2(DLY_TIME-1);

input  wire                 irst;         // reset signal
input  wire                 iclk;         // high speed sampling clock
input  wire                 xrst;         // reset signal
input  wire                 xclk;         // high speed sampling clock
output wire [DATA_BITS-1:0] xrcv_data;    // recieve data
output wire                 xrcv_rdy;     // receive data is ready
input  wire                 xrcv_ack;     // used to clear xrcv_rdy
input  wire       [OSW-1:0] din;          // BiPhase interface serial data input

// local registers
reg           [DS:0] tb_width;

reg        [OSL-1:0] rdata[0:OSW-1];
reg           [DC:0] data_cnt;
reg        [OSW-1:0] d3, d2, d1, d0;
reg        [OSW-1:0] last_polarity; // bit polarity at end of last data stream bit(s)
reg        [OSW-1:0] preamble_pol;  // polarity of preamble
reg                  DBrise;
reg                  DBfall;
reg            [2:0] TB_state;
reg            [2:0] TB_state_next;
reg           [DS:0] b_cnt;
reg           [DS:0] p1_len;  // length of time slots 0 - 2 in preamble 11100010 [or 00011101]
reg           [DS:0] p0_len;  // length of time slots 3 - 5 in preamble 11100010 [or 00011101]
reg           [DS:0] p10_len; // length of time slots 6 - 7 in preamble 11100010 [or 00011101]
reg                  rcv_flag;
reg  [DATA_BITS-1:0] idata;  // recieve data
reg                  irdy;
wire                 iack;
wire       [OSW-1:0] DB;
reg           [DS:0] tb_min, p1_min;
reg         [DS+1:0] tb_max, p1_max;
wire          [DS:0] slot_cnt_min;
wire                 data_capture_time;
wire                 data_stream_done;
reg                  preamble0_fail;
reg                  preamble10_fail;

localparam TB_IDLE         = 0,
           TB_PREAMBLE1    = 1,
           TB_PREAMBLE0    = 2,
           TB_PREAMBLE10   = 3,
           TB_DATA_STREAM  = 4;

generate
   if (ICLK_FREQ == XCLK_FREQ)
   begin: SAME_CLK
      assign xrcv_data = idata;
      assign xrcv_rdy  = irdy;
      assign iack      = xrcv_ack;
   end
   else // ICLK_FREQ != XCLK_FREQ
   begin: DIFF_CLK
      reg [DATA_BITS-1:0] xd1, xd0;
      reg xr2, xr1, xr0;
      reg ia1, ia0;

      assign xrcv_data = xd1;
      assign xrcv_rdy  = xr2;
      assign iack      = ia1;

      always @(posedge xclk) 
      begin
         {xd1, xd0} <= #TPD {xd0, idata};
         {xr2, xr1, xr0} <= #TPD {xr1, xr0, irdy};
      end

      always @(posedge iclk) 
      begin
         {ia1, ia0} <= #TPD {ia0, xrcv_ack};
      end
   end
endgenerate

generate
   genvar i;
   for (i = 0; i < OSW; i = i + 1)
   begin: STREAM_DATA
      always @(posedge iclk)
      begin
         if (rcv_flag)
            idata[i*OSL +: OSL] <= #TPD rdata[i];
 
         if (irst)
            rdata[i] <= #TPD 1'b0;
         else if (data_capture_time)
            rdata[i] <= #TPD {~(d3[i] ^ last_polarity[i]), rdata[i] [OSL-1:1]};   // if last polarity is the same as the curent polarity then it's a logic 1
      end
   end
endgenerate

always @(posedge iclk)
begin 
   if (irst)
      {d3, d2, d1, d0} <= #TPD {4*OSW{1'b0}};  // each dx is OSW bits wide
   else
      {d3, d2, d1, d0} <= #TPD {d2, d1, d0, din};

   if (irst)
      DBrise <= #TPD 1'b0;
   else
      DBrise <= #TPD d1[0] & !d2[0]; // rising edge of stream #0 DATA_BIT

   if (irst)
      DBfall <= #TPD 1'b0;
   else
      DBfall <= #TPD !d1[0] & d2[0]; // falling edge of stream #0 DATA_BIT

   if (irst || (TB_state == TB_IDLE))
      b_cnt <= #TPD 1'd0;
   else if (TB_state == TB_DATA_STREAM)
   begin
      if ((DBrise || DBfall) && slot_cnt_min)
         // maybe add code here Kirk: if b_cnt < tb_width then decrement tb_width, if b_cnt > tb_width then increment tb_width....adjust tb_width slightly
         b_cnt <= #TPD 1'd0;
      else
         b_cnt <= #TPD b_cnt + 1'd1;
   end

   if (irst || (TB_state == TB_IDLE))
      tb_width <= #TPD 0;
   else if (TB_state == TB_PREAMBLE10)
      tb_width <= #TPD ((p0_len+p1_len) >> 1); // average of p0_len and p1_len

   tb_min <= #TPD tb_width - (tb_width >> 1);
   tb_max <= #TPD tb_width + (tb_width >> 1);

   if (irst || (TB_state == TB_IDLE))
      p1_len <= #TPD 1'd0;
   else if (TB_state == TB_PREAMBLE1)
      p1_len <= #TPD p1_len + 1'd1;

   p1_min <= #TPD p1_len - (p1_len >> 1);
   p1_max <= #TPD p1_len + (p1_len >> 1);

   if (irst || (TB_state == TB_IDLE))
      p0_len <= #TPD 1'd0;
   else if (TB_state == TB_PREAMBLE0)
      p0_len <= #TPD p0_len + 1'd1;

   if (irst || (TB_state == TB_IDLE))
      preamble0_fail <= #TPD 1'd0;
   else if (TB_state == TB_PREAMBLE10)
      preamble0_fail <= #TPD (p0_len < p1_min) || (p0_len > p1_max);

   if (irst || (TB_state == TB_IDLE))
      p10_len <= #TPD 1'd0;
   else if (TB_state == TB_PREAMBLE10)
      p10_len <= #TPD p10_len + 1'd1;

   if (irst || (TB_state == TB_IDLE))
      preamble10_fail <= #TPD 1'd0;
   else if (TB_state == TB_DATA_STREAM)
      preamble10_fail <= #TPD (p10_len < tb_min) || (p10_len > tb_max);

   if (irst)
      preamble_pol <= #TPD {OSW{1'b0}};
   else if (TB_state == TB_IDLE)
      preamble_pol <= #TPD d3;

   if (irst)
      last_polarity <= #TPD {OSW{1'b0}};
   else if (TB_state == TB_IDLE)
      last_polarity <= #TPD preamble_pol;
   else if (data_capture_time)
      last_polarity <= #TPD d3;

   if (irst)
      TB_state <= #TPD TB_IDLE;
   else
      TB_state <= #TPD TB_state_next;

   rcv_flag <= #TPD ((TB_state == TB_DATA_STREAM) && data_stream_done);

   if (irst)
      irdy <= #TPD 1'b0;
   else if (rcv_flag)
      irdy <= #TPD 1'b1;
   else if (iack)
      irdy <= #TPD 1'b0;

   if (irst || (TB_state == TB_IDLE))
      data_cnt <= #TPD 1'b0;
   else if (data_capture_time)
      data_cnt <= #TPD data_cnt + 1'b1;
end
assign DB = d2;
assign data_stream_done = (data_cnt == OSL);
assign data_capture_time = (TB_state == TB_DATA_STREAM) && (DBrise || DBfall) && slot_cnt_min;
assign slot_cnt_min = (b_cnt >= tb_min);

// BiPhase receive state machine uses stream #0 for timing information and assumes streams #1 - (N-1) are in sync with stream #0
always @* 
begin
   case(TB_state)
      TB_IDLE:
      begin
         if ((DBrise && !preamble_pol[0]) || (DBfall & preamble_pol[0])) // wait for data input to go high/low depending on preamble_pol
            TB_state_next = TB_PREAMBLE1;
         else
            TB_state_next = TB_IDLE;  // wait for a DBrise to occur
      end

      TB_PREAMBLE1:  // record the high time (slots 0 - 2) of preamble pattern of 1 1 1 0 0 0 1 0 (or pattern 0 0 0 1 1 1 0 1)
      begin
         if (DB[0] ^ preamble_pol[0]) // only use data in first stream for state machine transitions
            TB_state_next = TB_PREAMBLE1;    // stay in this state while preamble is logic 1
         else
            TB_state_next = TB_PREAMBLE0;
      end

      TB_PREAMBLE0:  // record the low time (slots 3 - 5) preamble pattern of 1 1 1 0 0 0 1 0
      begin
         if (DB[0] ^ preamble_pol[0])
            TB_state_next = TB_PREAMBLE10;   // done when data bit goes high
         else
            TB_state_next = TB_PREAMBLE0;
      end

      TB_PREAMBLE10:  // now check last 2 slots of preamble pattern of 1 1 1 0 0 0 1 0
      begin
         if (preamble0_fail)
            TB_state_next = TB_IDLE;         // start over
         else if ((DBrise & !preamble_pol[0]) || (DBfall & preamble_pol[0]))
            TB_state_next = TB_DATA_STREAM;  // preamble finished and data stream is starting
         else
            TB_state_next = TB_PREAMBLE10;
      end

      TB_DATA_STREAM:
      begin
         if (preamble10_fail || data_stream_done) // add a data_stream fail which checks width of data bits
            TB_state_next = TB_IDLE; 
         else
            TB_state_next = TB_DATA_STREAM;  // continue getting data
      end

      default:
      begin
         TB_state_next = TB_IDLE;
      end
   endcase
end


function integer clogb2;
input [31:0] depth;
begin
   for (clogb2=0; depth>0; clogb2=clogb2+1)
      depth = depth >> 1;
end
endfunction

endmodule