/*
 BiPhase_xmit.v data transmitter - for use with BiPhase_rcv.v

 Kirk Weedman

 May 14, 2012
 
 Concept:
   Transmit data across N wires between FPGAs using a BiPhase Mark data encoding similar to the AES/SPDIF format

 BiPhase Mark Encoding:

   Data:        1   0   0   1   0   1   0
	           +-+ +---+   +-+ +---+ +-+   +-
	BPM:      -+ +-+   +---+ +-+   +-+ +---+
    
                1 1 1 0 0 0 1 0
              +-----+     +-+ +
	Preamble: -+     +-----+ +-+
   Preamble uses 8 time slots and each data bit 2 time slots. Depending on signal level before preamble begins, it may 
   transmit the 11100010 sequence or its inverse of 00011101.  After reset it will always output a low on dout and the
   very first sequence will be 11100010

 Method:
    DLY_TIME: time signal held low after reset occurs to allow matching BiPhase_rcv.v time to get ready to receive data stream
    OSL (Output Stream Length): Number of Data bits that will be transmitted in the data stream after the Preamble
    BIT_TIME: Amount of time it takes to transmit a Data "logic 0" or logic 1" (2 time slots)
	DEAD_TIME: time between transmissions. This should be set to the number of BIT_TIME periods desired

    After reset, the DOUT signal is held low for DLY_TIME, then the following is repeated:  Preamble, Data stream, DEAD_TIME ...

                 +---------------+ 
     data bit ---+               +-  "logic 0" if signal was low at end of last BIT_TIME
                 |<--BIT_TIME--->|
 
              ---+               +-
     data bit    +---------------+   "logic 0" if signal was high at end of last BIT_TIME
                 |<--BIT_TIME--->|

                 +-------+       +-  
     data bit ---+       +-------+   "logic 1" if signal was low at end of last BIT_TIME
                 |<--BIT_TIME--->|
 
              ---+       +-------+
     data bit    +-------+       +-  "logic 1" if signal was high at end of last BIT_TIME
                 |<--BIT_TIME--->|


                           +-----------------------+                +---
  DOUT --------------------+-----------------------+----------------+---
       |<---- DLY_TIME---->|<--- OSL*BIT_TIME ---->|<-- DEAD_TIME ->|
                           |<----------- 1/SEND_FREQ -------------->|



         ------------------------------------------+
xrdy                                               +--------------------------------------

                                              +---+---------------------------+
xreq     -------------------------------------+   +---------------------------+------

                                                +----------------------------+
xack     ---------------------------------------+                            +----------------

Note:   The transmitter is ready for more xdata ONLY when xrdy = 1
        xrdy initially goes high after reset once all the NUM_SYNC bits have been sent.
        xreq should only be asserted while xrdy = 1
        xreq should only be deasserted once xack = 1
        xreq MUST be removed before the xdata stream is completed (best to be low soon after xack goes low)
        xack will go low after the preamble is finished
*/

`timescale 1ns/100ps

module BiPhase_xmit (irst, iclk, xrst, xclk, xdata, xreq, xrdy, xack, dout);

parameter   OSW             = 1'b1;        // Output Stream Width
parameter   OSL             = 6'd32;       // Output Stream Length
parameter   XCLK_FREQ       = 48000000;    // Frequency in Hz of xclk
parameter   ICLK_FREQ       = 144000000;   // Frequency in Hz of iclk
parameter   DLY_TIME        = 200;         // delay time after reset (in milliseconds)
parameter   SEND_FREQ       = 400000;      // frequency in Hz of each complete OSL+DEAD_BITS transmission
parameter   DEAD_BITS       = 0;           // number of bits to remain inactive after OSL bits are sent
parameter   TPD             = 1;           // Typical Propagation Delay: only for simulation purposes
parameter   NUM_DLY_CLKS    = ICLK_FREQ/1000 * DLY_TIME;
localparam  BIT_TIME        = ICLK_FREQ/SEND_FREQ/(OSL+DEAD_BITS+4); // total clocks needed for each BIT_TIME time interval
localparam  SLOT_EVEN_CNT   = BIT_TIME/2;
localparam  SLOT_ODD_CNT    = BIT_TIME - SLOT_EVEN_CNT;  // BIT_TIME -> EVEN SLOT then ODD SLOT
parameter   DEAD_TIME       = DEAD_BITS*BIT_TIME;   // notice that the user can override this parameter and define the number
                                                    // of clocks directly for DEAD_TIME

localparam  DATAIN_BITS     = OSL * OSW;   // total data bits entered into xdata as {OSL_N-1, OSL_N-2...., OSL_0} where N = OSW
// example: OSL = 8, OSW = 3 so xdata is {OSL_2, OSL_1, OSL_0} which is 24 bits, where {OSL_2[0], OSL_1[0], OSL_0[0}} are the first
//          data bits transmitted in parallel.
localparam SS    = clogb2(OSL-1);
localparam DS    = clogb2(NUM_DLY_CLKS-1);
localparam QS    = clogb2(DEAD_TIME-1);

input  wire                   irst;  // reset signal for iclk domain
input  wire                   iclk;  // internal data sampling/state machine clock 
input  wire                   xrst;  // reset signal
input  wire                   xclk;  // external interface clock
input  wire [DATAIN_BITS-1:0] xdata;
input  wire                   xreq;  // someone wants to send new xdata
output wire                   xrdy;  // transmitter is ready for more xdata
output wire                   xack;  // acknowledge that xreq has occured
output reg          [OSW-1:0] dout;  // 1 to N bits output stream width

// local registers
reg              [2:0] NW_state;
reg             [DS:0] dly_cnt;
reg           [QS-1:0] b_cnt;
reg           [SS-1:0] stream_cnt;
reg                    irdy;
reg                    iack;
wire                   ireq;
reg          [OSL-1:0] id [0:OSW-1];
wire [DATAIN_BITS-1:0] idata;
wire         [OSW-1:0] OUT;
reg			     [2:0] slot_cnt;
wire             [7:0] preamble;
wire                   end_of_even_slot;
wire                   end_of_odd_slot;
wire                   end_of_slot;
wire                   preamble_done;
wire                   dly_time_done;
wire                   dead_time_done;
wire                   data_stream_done;
reg          [OSW-1:0] last_polarity;

localparam NW_IDLE         = 0,
           NW_ACK_WAIT     = 1,
           NW_PREAMBLE     = 2,
           NW_DATA_STREAM  = 3,
           NW_HOLD         = 4;

generate
if (ICLK_FREQ == XCLK_FREQ)
begin: SAME_CLK
   assign xrdy  = irdy;
   assign xack  = iack;

   assign idata = xdata;
   assign ireq  = xreq;
end
else
begin: DIFF_CLK
   reg xr1, xr0;
   reg xa1, xa0;
   reg [DATAIN_BITS-1:0] id1, id0;
   reg iq2, iq1, iq0;

   always @(posedge xclk) 
   begin
      {xr1, xr0} <= #TPD {xr0, irdy};
      {xa1, xa0} <= #TPD {xa0, iack};
   end

   always @(posedge iclk) 
   begin
      {id1, id0}  <= #TPD {id0, xdata};
      {iq2, iq1, iq0} <= #TPD {iq1, iq0, xreq}; // if xdata and xreq are at the same time make sure ireq will not occur before idata
   end

   assign xrdy = xr1;
   assign xack = xa1;

   assign idata = id1;
   assign ireq  = iq2;
end
endgenerate

always @(posedge iclk)
begin
   if (irst)
      irdy <= #TPD 1'b0;
   else if (NW_state == NW_ACK_WAIT)
      irdy <= #TPD 1'b1;
   else
      irdy <= #TPD 1'b0;

   if (irst)
      iack <= #TPD 1'b0;
   else if (irdy && ireq) // ignore requests until irdy is true
      iack <= #TPD 1'b1;
   else if ((NW_state == NW_PREAMBLE) && preamble_done) // wait till preamble finished before clearing xack
      iack <= #TPD 1'b0;

   if (irst)
      dly_cnt <= #TPD 1'b0;
   // else if (dly_cnt != NUM_DLY_CLKS) // Slow because of the size and causing timing analysis problems
   else if (!dly_cnt[DS]) // a little longer but the compare is much simpler for a large counter and FASTER
      dly_cnt <= #TPD dly_cnt + 1'b1;

   if (irst)
      b_cnt <= #TPD 1'b0;
   else if (NW_state == NW_ACK_WAIT)
      b_cnt <= #TPD 1'b0;
   else if ((NW_state == NW_PREAMBLE) && end_of_odd_slot) // it takes 2 "slots" to equa1 one BIT_TIME (first the EVEN slot then the ODD)
      b_cnt <= #TPD 1'b0;
   else if ((NW_state == NW_DATA_STREAM) && end_of_odd_slot) // it takes 2 "slots" to equa1 one BIT_TIME
      b_cnt <= #TPD 1'b0;
   else
      b_cnt <= #TPD b_cnt + 1'b1;

   if (irst || (NW_state == NW_ACK_WAIT))
      stream_cnt <= #TPD OSL-1;
   else if ((NW_state == NW_DATA_STREAM) && end_of_odd_slot)
      stream_cnt <= #TPD stream_cnt - 1'b1;

   if (irst || (NW_state == NW_ACK_WAIT))
      slot_cnt <= #TPD 7;
   else if ((NW_state == NW_PREAMBLE) && end_of_slot)
      slot_cnt <= #TPD slot_cnt - 1'd1;

   if (irst || (NW_state == NW_IDLE))
      last_polarity <= #TPD {OSW{1'b0}};
   else if (NW_state == NW_HOLD)
      last_polarity <= #TPD dout;
end
assign preamble = 8'b11100010;
assign end_of_even_slot = (b_cnt == SLOT_EVEN_CNT);
assign end_of_odd_slot = (b_cnt == BIT_TIME);
assign end_of_slot = end_of_even_slot || end_of_odd_slot;
assign preamble_done = (slot_cnt == 1'd0) && end_of_odd_slot;
assign dly_time_done = !dly_cnt[DS];
assign dead_time_done = (b_cnt == DEAD_TIME);
assign data_stream_done = (stream_cnt == 1'd0) && end_of_odd_slot;

generate
   genvar i;
   for (i = 0; i < OSW; i = i + 1)
   begin: STREAM_DATA
      always @(posedge iclk)
      begin
         if (NW_state == NW_ACK_WAIT)
            id[i] <= #TPD idata[i*OSL +: OSL];
         else if ((NW_state == NW_DATA_STREAM) && end_of_odd_slot)
            id[i] <= #TPD id[i] >> 1;  // LS bit is sent out first
      end
      assign OUT[i] = id[i][0];
   end
endgenerate

// BiPhase transmit state machine
always @(posedge iclk)
begin
   if (irst)
   begin
      NW_state <= #TPD NW_IDLE;
      dout <= #TPD {OSW{1'b0}};
   end
   else
   begin
      case(NW_state)
         NW_IDLE:
         begin
            dout <= #TPD {OSW{1'b0}};
            if (!dly_time_done)
               NW_state <= #TPD NW_IDLE;  // Wait at least DLY_TIME before sending OSL data bits
            else
               NW_state <= #TPD NW_ACK_WAIT;  // send preamble & OSL data bits once we get an ACK
         end

         NW_ACK_WAIT:
         begin
            if (iack)
            begin
               NW_state <= #TPD NW_PREAMBLE;  // we have permission to transmit Preamble and then a stream of OSL data bits
               dout <= #TPD {OSW{preamble[7]}} ^ last_polarity; // send first preamble bit to dout signal(s)
            end
            else
            begin
               NW_state <= #TPD NW_ACK_WAIT;
               dout <= #TPD last_polarity; // force dout to last_polarity till we get the ack
            end
         end

         NW_PREAMBLE: // preamble of 11100010 = 8 time slots = 4 data bits wide [i.e. 4 BIT_TIME wide]
         begin
            if (preamble_done)// 8 slots for preamble
            begin
               NW_state <= #TPD NW_DATA_STREAM;
               dout <= #TPD ~dout; // bit toggle
            end
            else if (end_of_slot)
            begin
               NW_state <= #TPD NW_PREAMBLE;
               dout <= #TPD {OSW{preamble[slot_cnt-1]}} ^ last_polarity; // slot_cnt = 7,6,...0
            end
         end

         NW_DATA_STREAM: // now send a stream of OSL data bits in BiPhase Mark Encoding
         begin
            if (data_stream_done)
            begin
               NW_state <= #TPD NW_HOLD;
               dout <= #TPD ~dout; // always change polarity
            end
            else if (end_of_even_slot)
               dout <= #TPD dout ^ OUT; // if data bit being sent is a logic 1 then polarity changes halfway through BIT_TIME
            else if (end_of_odd_slot)
               dout <= #TPD ~dout; // always change polarity
         end

         NW_HOLD: // send out a low signal during this time
         begin
            dout <= #TPD dout;
            if (!dead_time_done)
               NW_state <= #TPD NW_HOLD;
            else
               NW_state <= #TPD NW_ACK_WAIT;      // DEAD_BITS time has finished
         end

         default:
         begin
            dout <= #TPD {OSW{1'b0}};
            NW_state <= #TPD NW_IDLE;
         end
      endcase
   end
end

function integer clogb2;
input [31:0] depth;
begin
   for(clogb2=0; depth>0; clogb2=clogb2+1)
      depth = depth >> 1;
end
endfunction

endmodule